from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import object
from builtins import super
from future import standard_library
standard_library.install_aliases()

from valvevmf.parser import VmfParse
from valvevmf.writer import VmfWrite


class Vmf(object):
    """
    This is the basic class to interact with vmf files, it is mostly a collection of VmfNodes.
    """

    def __init__(self, path=None):
        """
        initalize a Vmf file.

        :param path: The location of the vmf file to be parsed, saved as :any:`source_path<Vmf.source_path>`.
        :type path: str, optional
        """

        #: :type: (str) - The location of the parsed file
        self.source_path = path

        #: :type: (list[VmfNode]) - The list of nodes at the Vmf root
        self.nodes = []

        if self.source_path:
            self.nodes = VmfParse(self.source_path)

    def save(self, destination=None):
        """Saves the current instance of the Vmf. Overwrites original vmf file if no destination is provided.

        :param destination: A path (directory + filename) to determine where to save the vmf file.
        :type destination: str, optional
        """
        VmfWrite(self, destination or self.source_path)

    def vmf_str(self):
        vstr = ''
        for n in self.nodes:
            vstr += n.vmf_str()
        return vstr



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
from future import standard_library
standard_library.install_aliases()

from decimal import Decimal


def _repr_prop_value(val, wrapper=None):
    vstr = val
    if isinstance(val, str):
        vstr = val
    elif isinstance(val, Decimal):
        vstr = '{:.6g}'.format(val)
    elif isinstance(val, bool):
        vstr = '1' if val else '0'
    elif isinstance(val, int):
        vstr = str(val)
    elif isinstance(val, list) or isinstance(val, tuple):
        vstr = ' '.join([_repr_prop_value(c) for c in val])

    if isinstance(wrapper, str) and len(wrapper):
        return wrapper[0] + vstr + wrapper[-1]
    return vstr


def _property_str(key, val, indent):
    return '   ' * indent + '"' + key + '" "' + val + '"' + '\n'


def write_property(prop, node_name=None, indent=0):

    key, pvalue = prop
    if (node_name == 'editor' and key == 'logicalpos') or \
       (node_name == 'dispinfo' and key == 'startposition'):
        pvalue = _repr_prop_value(pvalue, '[]')
    elif node_name == 'cordon' and key in ['mins', 'maxs']:
        pvalue = _repr_prop_value(pvalue, '()')
    elif node_name == 'side' and key in ['uaxis', 'vaxis']:
        pvalue = _repr_prop_value(pvalue[:4], '[]') + \
            ' ' + _repr_prop_value(pvalue[4])
    elif node_name == 'side' and key == 'plane':
        pvalue = ' '.join([_repr_prop_value(c, '()') for c in pvalue])
    elif node_name in ['normals', 'distances', 'offsets', 'offset_normals',
                       'alphas', 'triangle_tags'] and key == 'rows':
        rows = [('row' + str(index), _repr_prop_value(r))
                for index, r in enumerate(pvalue)]
        return ''.join([_property_str(r[0], r[1], indent) for r in rows])
    else:
        pvalue = _repr_prop_value(pvalue)

    return _property_str(key, pvalue, indent)



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import open
from future import standard_library
standard_library.install_aliases()


def VmfWrite(vmf, filename):

    vmf_text = vmf.vmf_str()

    f = open(filename, "w")
    f.write(vmf_text)
    f.close()



from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from builtins import map
from builtins import str
from builtins import range
from builtins import dict
from builtins import open
from builtins import int
from future import standard_library
standard_library.install_aliases()

from collections import namedtuple

import os  # NOQA: E402
from pyparsing import *  # NOQA: E402

from valvevmf.vmf import *  # NOQA: E402
from valvevmf.node import VmfNode  # NOQA: E402

from valvevmf.property_parser import parse_property, pp_float, pp_uint8, pp_vertex_content


def asTuple(data):
    return [tuple(data.asList()[0])]


def asList(data):
    return list(data.asList())


def asGrid(data):
    return ('rows', [d for d in data])


# Node parsing
pp_vnode = Forward()
pp_node_name = Word(alphas+'_')
pp_node_keyval = Group(QuotedString('"') + QuotedString('"'))
pp_node_keyval.setParseAction(asTuple)

pp_gridnode_float_name = Literal('distances') ^ Literal('alphas')
pp_gridnode_float_cont = Suppress('{') + ZeroOrMore(
    Suppress(QuotedString('"')) + Suppress('"') +
    Group(ZeroOrMore(pp_float)).setParseAction(asList) +
    Suppress('"')).setParseAction(asGrid) + Suppress('}')

pp_gridnode_uint8_name = Literal('triangle_tags')
pp_gridnode_uint8_cont = Suppress('{') + ZeroOrMore(
    Suppress(QuotedString('"')) + Suppress('"') +
    Group(ZeroOrMore(pp_uint8)).setParseAction(asList) +
    Suppress('"')).setParseAction(asGrid) + Suppress('}')

pp_gridnode_vertex_name = Literal('normals') ^ Literal('offsets') ^ \
    Literal('offset_normals')
pp_gridnode_vertex_cont = Suppress('{') + ZeroOrMore(
    Suppress(QuotedString('"')) + Suppress('"') +
    Group(ZeroOrMore(pp_vertex_content)).setParseAction(asList) +
    Suppress('"')).setParseAction(asGrid) + Suppress('}')

pp_gridnode_float = Group(pp_gridnode_float_name + pp_gridnode_float_cont)
pp_gridnode_uint8 = Group(pp_gridnode_uint8_name + pp_gridnode_uint8_cont)
pp_gridnode_vertex = Group(pp_gridnode_vertex_name + pp_gridnode_vertex_cont)

pp_node_contents = Suppress('{') + \
    ZeroOrMore(pp_node_keyval ^ pp_gridnode_float ^
               pp_gridnode_uint8 ^ pp_gridnode_vertex ^
               pp_vnode).setParseAction(asList) + \
    Suppress('}')

pp_vnode <<= Group(pp_node_name + pp_node_contents)

pp_vmf = OneOrMore(pp_vnode)


def VmfParse(filename):
    results = []
    try:
        results = pp_vmf.parseFile(filename, encoding="iso-8859-1")
    except Exception as e:
        raise

    def make_node(node_result):
        name = node_result.pop(0)
        properties = [parse_property(p, name)
                      for p in node_result if isinstance(p, tuple)]
        subnodes = [make_node(n)
                    for n in node_result if isinstance(n, list)]
        node = VmfNode(name, properties, subnodes)
        return node

    return [make_node(r) for r in results]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals
from builtins import str
from future import standard_library
standard_library.install_aliases()
from builtins import object

from valvevmf.property_writer import write_property


class VmfNode(object):
    def __init__(self, name, properties=None, nodes=None):
        """Creates an empty instance of VmfNode.

        :param name: indicates the node's type
        :type name: str
        :param properties: The node's properties.
        :type properties: list[tuple], optional
        :param nodes: The node's sub-nodes
        :type nodes: list[VmfNode], optional
        """

        if properties is None:
            properties = []

        if nodes is None:
            nodes = []

        #: :type: (str) - The nodes's name ("world", "solid", "entity"...)
        self.name = name
        #: :type: (list[tuple])
        self.properties = properties
        #: :type: (list[VmfNode], optional)
        self.nodes = nodes

    def __repr__(self):
        """A partial, printable summary of a VmfNode.

        :returns: A Python formated string.
        :rtype: str
        """

        return '<VmfNode %(name)s %(p_len)x %(n_len)x' % \
            {'name': self.name,
             'p_len': len(self.properties),
             'n_len': len(self.nodes)}

    def vmf_str(self, indent=0):
        vstr = '    ' * indent + self.name + '\n'
        vstr += '    ' * indent + '{\n'
        for p in self.properties:
            vstr += write_property(p, self.name, indent+1)
        for n in self.nodes:
            vstr += n.vmf_str(indent+1)
        vstr += '     ' * indent + '}\n'
        return vstr



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import int
from future import standard_library
standard_library.install_aliases()

from pyparsing import *  # NOQA: E402
from decimal import Decimal  # NOQA: E402


def asTuple(data):
    return [tuple(data.asList()[0])]


# Property parsing
pp_bool = Word(nums+'-').setParseAction(lambda p: bool(int(p[0])))
pp_int = Word(nums+'-').setParseAction(lambda p: int(p[0]))
pp_uint8 = Word(nums).setParseAction(lambda p: min(255, max(0, int(p[0]))))
pp_float = Word(nums+'-.').setParseAction(lambda p: Decimal(p[0]))

pp_angle = Group(pp_float + pp_float + pp_float).setParseAction(asTuple)
pp_origin = pp_angle

pp_vertex_content = Group(pp_float + pp_float +
                          pp_float).setParseAction(asTuple)
pp_vertex_tup = Suppress('(') + pp_vertex_content + Suppress(')')
pp_vertex_arr = Suppress('[') + pp_vertex_content + Suppress(']')

pp_plane = Group(pp_vertex_tup + pp_vertex_tup +
                 pp_vertex_tup).setParseAction(asTuple)
pp_color255 = Group(pp_uint8 + pp_uint8 + pp_uint8).setParseAction(asTuple)

pp_uvaxis = Group(Suppress('[') + pp_float + pp_float + pp_float + pp_float +
                  Suppress(']') + pp_float).setParseAction(asTuple)
pp_2dvector = Group(Suppress('[') + pp_float + pp_float +
                    Suppress(']')).setParseAction(asTuple)

pp_allowed_verts = Group(ZeroOrMore(pp_int)).setParseAction(asTuple)

param_type_map = {
    'id': pp_int,
    'look': pp_vertex_arr,

    'entity': {
        'classname': None,
        'targetname': None,
        'comment': None,
        'angles': pp_angle,
        'origin': pp_origin,

        'position': pp_vertex_arr,
    },
    'versioninfo': {
        'prefab': pp_bool,
        'editorversion': pp_int,
        'editorbuild': pp_int,
        'mapversion': pp_int,
        'formatversion': pp_int,
    },
    'viewsettings': {
        'bSnapToGrid': pp_bool,
        'bShowGrid': pp_bool,
        'bShowLogicalGrid': pp_bool,
        'bShow3DGrid': pp_bool,
        'nGridSpacing': pp_int,
    },
    "world": {
        'maxpropscreenwidth': pp_int,
        'detailmaterial': None,
        'detailvbsp': None,
        'skyname': None,
    },
    'visgroup': {
        'name': None,
        'color': pp_color255,
        'visgroupid': pp_int,
    },
    'editor': {
        'color': pp_color255,
        'groupid': pp_int,
        'visgroupid': pp_int,
        'visgroupshown': pp_bool,
        'visgroupautoshown': pp_bool,
        'logicalpos': pp_2dvector,

    },
    'side': {
        'plane': pp_plane,
        'material': None,
        'uaxis': pp_uvaxis,
        'vaxis': pp_uvaxis,
        'rotation': pp_float,
        'lightmapscale': pp_int,
        'smoothing_groups': pp_int,
    },
    'dispinfo': {
        'power': pp_int,
        'startposition': pp_vertex_arr,
        'flags': pp_int,
        'elevation': pp_int,
        'subdiv': pp_bool,
    },
    'allowed_verts': {
        '10': pp_allowed_verts,
    },
    'cameras': {
        'activecamera': pp_int,
    },
    'cordon': {
        'active': pp_bool,
        'mins': pp_vertex_tup,
        'maxs': pp_vertex_tup,
    }
}


def parse_property(keyvalue, node_name=None):
    key, value = keyvalue
    target_dict = param_type_map
    prop_parser = target_dict.get(key, None)
    if node_name:
        target_dict = target_dict.get(node_name, target_dict)
    prop_parser = target_dict.get(key, prop_parser)

    if prop_parser:
        value = prop_parser.parse_string(value)[0]
    return (key, value)
